/*
 * t-bots.c
 *
 * toOl robot
 *
 */

#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "RobSock.h"

#include "t-bots.h"

static int final_time, key_time;
static struct beaconMeasure bcn;

char name[20] = "t-bot";

char fname[20];
FILE *outf;

/* this is unique for each robot in the range [0, n_robots - 1]*/
int id = 0;

#define MAX_DESTS  (8)

int dest_id = 0;
float dest[MAX_DESTS][2] = {{5,5},{5,10},{14,5},{14,10},{23,5},{23,10},{14,5},{14,10}};

int init_dest_id[5] = {0, 1, 2, 4, 5};

struct t_status statuses[5];

float min_dist = DIST_INFTY;

int sm = 2;   /* current speed mode */

//struct t_speed spd[7] = {{0.01, 30, 150}, {0.03, 30, 150}, {0.06, 30, 150}, {0.07, 30, 150}, {0.09, 30, 150}, {0.11, 30, 150}, {0.13, 30, 150}};
struct t_speed spd[7] = {{0.01, 50, 150}, {0.03, 50, 150}, {0.06, 50, 150}, {0.07, 50, 150}, {0.09, 50, 150}, {0.10, 50, 150}, {0.11, 50, 150}};

struct t_bf_info bf_info;
struct t_dist_hist dist_hist;
struct t_pos_hist pos_hist;
struct t_bcn_dir_hist bcn_dir;

float get_dir(struct t_status *s)
{
  float sum = 0.0;
  int i, n = 0;

  if (s->full) {
    for (i = s->idir; i < NDIRS; i++, n++) sum += s->dir[i];
  }

  for (i = 0; i < s->idir; i++, n++) sum += s->dir[i];

  return sum / n;
}

void set_dir(struct t_status *s, float d)
{
  s->dir[s->idir] = d;
  if (s->idir++ >= NDIRS) {
    s->idir = 0;
    s->full = 1;
  }
}

/************************************************************
 *
 * Functions to manage the linear regression of distances.
 *
 ************************************************************/

void reset_dist_hist()
{
  dist_hist.full = 0;
  dist_hist.i = 0;
}

void add_dist_measure(float d)
{
  dist_hist.dist[dist_hist.i] = d;
  dist_hist.i++;
  if (dist_hist.i >= DIST_HIST_MAX) {
    dist_hist.full = 1;
    dist_hist.i = 0;
  }
}

int dist_ready()
{
  return dist_hist.full;
}

/*
 * The linear regression is calculated as
 * $m = \frac{n \sum x y - \sum x \sum y}{n (\sum x^2) - (\sum x)^2}$
 */
float dist_regr_slope()
{
  float sumx = 0.0, sumx2 = 0.0, sumy = 0.0, sumxy = 0.0;
  float x = 0.0;
  float m;
  int i;

  if (dist_hist.full) {
    for (i = dist_hist.i; i < DIST_HIST_MAX; i++) {
      sumx += x;
      sumx2 += x * x;
      sumy += dist_hist.dist[i];
      sumxy += x * dist_hist.dist[i];
      x++;
    }
    for (i = 0; i < dist_hist.i; i++) {
      sumx += x;
      sumx2 += x * x;
      sumy += dist_hist.dist[i];
      sumxy += x * dist_hist.dist[i];
      x++;
    }
    m = (DIST_HIST_MAX * sumxy - sumx * sumy) / 
        (DIST_HIST_MAX * sumx2 - sumx * sumx);
  } else {
    fprintf(outf, "%s(%d): %s\n", __FILE__, __LINE__, __FUNCTION__);
    exit(1);
    /* TODO: management of this condition; but should not happen in
 		 * normal working condition 
     */
  }

  return m;
}

void add_pos_measure(float x, float y)
{
  pos_hist.x[pos_hist.i] = x;
  pos_hist.y[pos_hist.i] = y;
  pos_hist.i++;
  if (pos_hist.i > TMAX) {
    fprintf(outf, "%s(%d): %s\n", __FILE__, __LINE__, __FUNCTION__);
    exit(1);
    /* TODO: management of this condition; but should not happen in
 		 * normal working condition 
     */
  }
}

/*
 * Returns the center of mass evaluated over the last "back" positions.
 */
int get_pos_center(int back, float *x, float *y)
{
  int n;
  float sumx = 0.0, sumy = 0.0;

  if (back <= 0) return -1;

  n = back;
  if (back > pos_hist.i) n = pos_hist.i;

  for (back = n; back >= 0; back--) {
    sumx += pos_hist.x[pos_hist.i - back];
    sumy += pos_hist.y[pos_hist.i - back];
  }
  *x = sumx / n;
  *y = sumy / n;

  return 0;
}

/************************************************************
 *
 * Some useful math functions.
 *
 ************************************************************/

float rad2deg(float rad)
{
  return 180.0 * rad / M_PI;
}

float deg2rad(float deg)
{
  return deg * M_PI / 180.0;
}

float dist(float x1, float y1, float x2, float y2)
{
  return sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2));
}

float head(float x1, float y1, float x2, float y2)
{
  return rad2deg(atan2((y2 - y1), (x2 - x1)));
}

float diff_ang2(float a1, float a2)
{

  /* rotate a1 with a2 so that the sought after
   * angle is between the resulting angle and the x-axis */
  a1 -= a2;

  /* "normalize" a1 to range [-180,180) */
  while(a1 < -180)
    a1 += 360;
  while(a1 >= 180)
    a1 -= 360;

  /* a1 has the signed answer */
  return -a1;
}

/* the result of diff_ang2 has the signed answer, just "unsign it" */
float diff_ang(float a1, float a2)
{
  return fabs(diff_ang2(a1, a2));
}

float norm(float a)
{
  return fmod(fmod(a, 360.0) + 360.0, 360.0);
}

/************************************************************
 *
 * Some functions to manage the communication.
 *
 ************************************************************/

void beacon_found(struct t_status *s, float dir)
{
  // keeps only the first detection (due to the approaching strategy)
  // CAUTION: not sure about this!
  // A better stategy would be to register all the measurements and 
  // use them to improve the estimation (easy to do!!)
  if (s->record_new_beacon) {
    s->xc = s->x;
    s->yc = s->y;
    s->dir2bcn = fmod(get_dir(s) + dir + 720.0, 360.0);
    s->mask |= BEACON;
    s->mask |= (0x01 << id);
    s->record_new_beacon = 0;
  }
}

/************************************************************
 *
 * Functions to be classified. :-)
 *
 ************************************************************/

bool get_gps(struct t_status *s)
{
  if ((s->gps_rdy = IsGPSReady())) {
    s->t = GetTime();
    s->x = GetX();
    s->y = GetY();
    set_dir(s, GetDir());
  }

  return s->gps_rdy;
}

/*
 * Determines the motor power with a proportional control to rotate
 * the robot towards the desired direction. The movement is
 * performed along the directions that leads to the shortes
 * rotation.
 * DC_power represent the constant value summed to the differential
 * control values to achieve the precession of the robot (if 0 the 
 * robot simply rotates).
 * The gain controls the amplitude of the control.
 * The th value is a threshold value that impose a rotation towards
 * a constant direction if the absolute error is within the
 * threshold value. This should allow to avoid unwanted stallings when
 * errors on the direction makes the control to oscillate.
 */

void adjust_angle(struct t_status *s, 
                  float dir, float dir_sp,
                  float DC_power, float gain, float th)
{
  float diff_power, diff_ang;

  diff_ang = diff_ang2(dir_sp, dir);

  /* takes into account the errors on the direction sampling */
  if (fabs(diff_ang) < th) diff_ang = fabs(diff_ang);
  diff_power = gain * diff_ang;

  s->lPow = DC_power - diff_power / 2;
  s->rPow = DC_power + diff_power / 2;
}

int go_to_point(struct t_status *s, 
                struct beaconMeasure bcn)
{
  float dir;

  dir = rad2deg(atan2(s->goaly - s->y, s->goalx - s->x));

  if (s->cx > 3.2 || s->dx > 3.2 || s->sx > 3.2 || s->hit) {
    if (s->sx > s->dx) {
      s->lPow = 0.9 * spd[sm].max;
      s->rPow = -0.9 * spd[sm].max;
    } else {
      s->lPow = -0.9 * spd[sm].max;
      s->rPow = 0.9 * spd[sm].max;
    }
  } else if (s->dx > 1.5) { /* obstacle near: avoid */
    s->lPow = 0.0 * spd[sm].max;
    s->rPow = 0.6 * spd[sm].max;
  } else if (s->sx > 1.5) {
    s->lPow = 0.6 * spd[sm].max;
    s->rPow = 0.0 * spd[sm].max;
  } else { 
    adjust_angle(s, dir, get_dir(s), 0.9 * spd[sm].max, 0.001, 0.0);
  }

  return 0;
}

void motion_to_goal(struct t_status *s, 
                float d,
                struct beaconMeasure bcn)
{
  go_to_point(s, bcn);
}

void boundary_following(struct t_status *s,
                float d,
                struct beaconMeasure bcn)
{
  static int timer = 0;

  /* close to obstacle: rotate */
  if (s->cx > 3.5 || s->dx > 3.5 || s->sx > 3.5 || s->hit) {
     s->lPow = 0.9 * spd[sm].max;
     s->rPow = -0.9 * spd[sm].max;
     timer = 0;
  } else {
    if (s->sx > 1.5) { /* go along the obstacle */
      s->lPow = 0.9 * spd[sm].max;
      s->rPow = 0.9 * spd[sm].max;
      timer = 0;
    } else {
      if (timer > spd[sm].inf_loop_timer) {   /* avoids infinite loops */
        s->lPow = 0.9 * spd[sm].max;
        s->rPow = 0.9 * spd[sm].max;
      } else {
        s->lPow = -0.5 * spd[sm].max;
        s->rPow = 0.9 * spd[sm].max;
        timer++;
      }
    }
  }
}

/*
 * When a new motion-to-goal state is entered, we need to reset the
 * history of distances and the min_dist value, then set the new
 * FSM's state.
 */
void new_motion2goal(struct t_status *s, int next)
{
  reset_dist_hist();
  min_dist = DIST_INFTY;
  s->fsm = next;
  s->m_fsm = MOTION_TO_GOAL;
}

void set_goal(struct t_status *s, float x, float y)
{
  if (x < 0.0) x = 0.0;
  if (x > 28.0) x = 28.0;
  
  if (y < 0.0) y = 0.0;
  if (y > 14.0) y = 14.0;
  
  s->goalx = x;
  s->goaly = y;
}

int check_watchdog(struct t_status *s)
{
  /* check redundant positions after the timer has expired */
  if (s->t - bf_info.t > spd[sm].pos_recurr_timer) {
    if (dist(s->x, s->y, bf_info.x, bf_info.y) < 0.8) {
      bf_info.t = 100000;
      return 1;
    }
  }
  return 0;
}

int stalled(struct t_status *s)
{
  float x, y;

  get_pos_center(200, &x, &y);
  if (dist(s->x, s->y, x, y) < 0.5)
    return 1;
  
  return 0;
}

int check_try_enter(struct t_status *s)
{
  if (is_proximity(s, 1.8) && !s->gnd && how_many_onground() >= 4) {
    return 1;
  }
  return 0;
}

float head_to_goal(struct t_status *s, struct beaconMeasure bcn)
{
  float d1, d2;

  d1 = head(s->x, s->y, s->goalx, s->goaly);
  d2 = diff_ang2(d1, get_dir(s));
  return norm(d2);
}

void quest(struct t_status *s,
           float d,
           struct beaconMeasure bcn)
{
  float d1;
  switch (s->m_fsm) {
    case MOTION_TO_GOAL : 
      if (s->mask & 0x1f00) {   /* one robot is enough */
        new_motion2goal(s, GROUND_FOUND);
        break;
      }
      d1 = head_to_goal(s, bcn); 
      if (d1 < 180.0) {
        s->m_fsm = MOTION_TO_GOAL;
      } else {
        s->m_fsm = BOUNDARY_FOLLOWING;
        bf_info.x = s->x;
        bf_info.y = s->y;
        bf_info.t = s->t;
      }
      if (d < min_dist) min_dist = d;
      if (min_dist < 0.8) s->m_fsm = ON_GOAL;
      if (s->mask & 0x1f) {   /* one robot is enough */
        new_motion2goal(s, DIR_FOUND);
      }
      set_goal(s, dest[dest_id][0], dest[dest_id][1]);
      motion_to_goal(s, d, bcn);
      break;

    case BOUNDARY_FOLLOWING :
      s->record_new_beacon = 1;   /* not sure : maybe cause unstability? */
      if (s->mask & 0x1f00) {   /* one robot is enough */
        new_motion2goal(s, GROUND_FOUND);
        break;
      }

      if (d < min_dist) {
        min_dist = d;
        s->m_fsm = MOTION_TO_GOAL;
      }
      if (min_dist < 0.8) s->m_fsm = ON_GOAL;
      if (check_watchdog(s))
        s->m_fsm = ON_GOAL;
      if (s->mask & 0x1f) {   /* one robot is enough */
        new_motion2goal(s, DIR_FOUND);
      }
      set_goal(s, dest[dest_id][0], dest[dest_id][1]);
      boundary_following(s, d, bcn);
      break;

    case ON_GOAL :
      dest_id = (dest_id + 1) % MAX_DESTS;
      new_motion2goal(s, QUEST);
      break;
  }
}

void ground_found(struct t_status *s,
                float d,
                struct beaconMeasure bcn)
{
  int lid;
  float d1;

  switch (s->m_fsm) {
    case MOTION_TO_GOAL : 
      s->record_new_beacon = 1;   /* not sure!!! */
      d1 = head_to_goal(s, bcn); 
      if (d1 < 180.0) {
        s->m_fsm = MOTION_TO_GOAL;
      } else {
        s->m_fsm = BOUNDARY_FOLLOWING;
        bf_info.x = s->x;
        bf_info.y = s->y;
        bf_info.t = s->t;
      }
      if (d < min_dist) min_dist = d;
      if (min_dist < 0.8) s->m_fsm = ON_GOAL;
      lid = get_leader_id(s);
      if (lid != -1) {
        set_goal(s, statuses[lid].x, statuses[lid].y);
      } else {
        if (s->mask & TARGET) {
          set_goal(s, s->xgnd, s->ygnd);
        } else {
          set_goal(s, s->xc + 4 * cos(deg2rad(s->dir2bcn)), s->yc + 4 * sin(deg2rad(s->dir2bcn)));
        }
      }
      motion_to_goal(s, d, bcn);
      break;

    case BOUNDARY_FOLLOWING :
      s->record_new_beacon = 1;   /* not sure!!! */
      if (d < min_dist) {
        min_dist = d;
        s->m_fsm = MOTION_TO_GOAL;
      }
      if (min_dist < 0.8) s->m_fsm = ON_GOAL;
      if (check_watchdog(s))
        s->m_fsm = ON_GOAL;
      if (s->mask & TARGET) {
        set_goal(s, s->xgnd, s->ygnd);
      } else {
        set_goal(s, s->xc + 4 * cos(deg2rad(s->dir2bcn)), s->yc + 4 * sin(deg2rad(s->dir2bcn)));
      }
      boundary_following(s, d, bcn);
      break;

    case ON_GOAL :
      new_motion2goal(s, GROUND_FOUND);
      break;
  }
}

void dir_found(struct t_status *s,
               float d,
               struct beaconMeasure bcn)
{
  float d1;

  switch (s->m_fsm) {
    case MOTION_TO_GOAL : 
      if (s->mask & 0x1f00) {   /* one robot is enough */
        new_motion2goal(s, GROUND_FOUND);
        break;
      }

      d1 = head_to_goal(s, bcn); 
      if (d1 < 180.0) {
        s->m_fsm = MOTION_TO_GOAL;
      } else {
        s->m_fsm = BOUNDARY_FOLLOWING;
        bf_info.x = s->x;
        bf_info.y = s->y;
        bf_info.t = s->t;
      }
      if (d < min_dist) min_dist = d;
      if (min_dist < 0.8) s->m_fsm = ON_GOAL;

      set_goal(s, 
               s->xc + 6 * cos(deg2rad(s->dir2bcn)),
               s->yc + 6 * sin(deg2rad(s->dir2bcn)));
      motion_to_goal(s, d, bcn);
      break;

    case BOUNDARY_FOLLOWING :
      s->record_new_beacon = 1;   /* not sure : maybe cause unstability!!! */
      if (s->mask & 0x1f00) {   /* one robot is enough */
        new_motion2goal(s, GROUND_FOUND);
        break;
      }

      if (d < min_dist) {
        min_dist = d;
        s->m_fsm = MOTION_TO_GOAL;
      }
      if (min_dist < 0.8) s->m_fsm = ON_GOAL;
      set_goal(s,
               s->xc + 4 * cos(deg2rad(s->dir2bcn)),
               s->yc + 4 * sin(deg2rad(s->dir2bcn)));
      boundary_following(s, d, bcn);
      if (check_watchdog(s))
        s->m_fsm = MOTION_TO_GOAL;
      break;

    case ON_GOAL :
      s->record_new_beacon = 1;
      new_motion2goal(s, DIR_FOUND);
      break;
  }
}

int is_onground(struct t_status *s1)
{
  if (s1->gnd) {
    return 1;
  }
  return 0;
}

int how_many_onground()
{
  int i, num = 0;

  for (i = 0; i < NROBOT; i++) {
    if (is_onground(&statuses[i])) {
      num++;
    }
  }
  return num;
}

int is_proximity(struct t_status *s, float dth)
{
  int i, lid;
  float d;

  lid = get_leader_id(s);
  if (lid == -1) return 0;
  for (i = 0; i < NROBOT; i++) {
    d = dist(s->x, s->y, statuses[lid].xgnd, statuses[lid].ygnd);
    if (d < dth) {
      return 1;
    }
  }
  return 0;
}

void check_can_finish(struct t_status *s)
{
  if (((s->mask & 0x1f00) == 0x1f00) && s->gnd && how_many_onground() == NROBOT)
    s->fsm = CAN_FINISH;
}

int get_leader_id(struct t_status *s)
{
  int i;

  for (i = 0; i < NROBOT; i++) {
    if (statuses[i].mask & LEADER) {
      return i;
    }
  }
  return -1;
}

int get_try_entering_id(struct t_status *s)
{
  int i;

  for (i = 0; i < NROBOT; i++) {
    if (statuses[i].fsm == TRY_ENTER ||
        statuses[i].fsm == TRY_ENTER_ALIGNED) {
      return i;
    }
  }
  return -1;
}

int check_someone_finding_center(struct t_status *s)
{
  int i;

  for (i = 0; i < NROBOT; i++) {
    if (i != id) {
      if ((statuses[i].m_fsm == ONGROUND_ROT ||
          statuses[i].m_fsm == ONGROUND_GTB) &&
          statuses[i].fsm == ON_GROUND &&
          dist(s->x, s->y, statuses[i].x, statuses[i].y) < 5.0
         ) {
        return i;
      }
    }
  }
  return 0;
}

int get_avg_pos(int n, float *x, float *y)
{
  int i;

  *x = 0.0;
  *y = 0.0;

  if (n <= 0) return -1;
  if (n > pos_hist.i) n = pos_hist.i;
  for (i = 0; i < n; i++) {
    *x += pos_hist.x[pos_hist.i - i];
    *y += pos_hist.y[pos_hist.i - i];
  }
  *x /= n;
  *y /= n;
  return 1;
}

int aligned_id(struct t_status *s)
{
  int i;

  for (i = 0; i < NROBOT; i++) {
    if (i != id) {
      if (statuses[i].fsm == TRY_ENTER_ALIGNED) {
        return i;
      }
    }
  }
  return -1;
}

/*
 * Returns the number of onground movable robots, aand an ordered
 * array of id's that start with the closer (supplied as cl),
 * ordered ascending wih respect to the rotation direction dir (that
 * could be positive or negative depending on the rotation direction
 * of the closer.
 */
int get_onground_movable(struct t_status *s, int *v)
{
  int i;
  int mov = 0;

  for (i = 0; i < NROBOT; i++) {
    if ((statuses[i].fsm == ON_GROUND) && !(statuses[i].mask & LEADER)) {
      v[mov] = i;
      mov++;
    }
  }
  return mov;
}

void swap_int(int *a, int *b)
{
  int tmp;

  tmp = *a;
  *a = *b;
  *b = tmp;
}

void get_final_ang(struct t_status *s, float *a)
{
  int i, j;
  float n1, n2;
  int cid, eid;
  int v[NROBOT];
  int nmov;
  float ang[3] = {90.0, 180.0, 270.0};
  float da;

  cid = get_closer(s, &da);
  eid = aligned_id(s);
  nmov = get_onground_movable(s, v);
  da = diff_ang2(statuses[eid].bcn_dir, statuses[cid].bcn_dir);

  /* order the vector */
  for (i = 0; i < nmov; i++) {
    for (j = i + 1; j < nmov; j++) {
      n1 = norm(diff_ang2(statuses[eid].bcn_dir, statuses[v[i]].bcn_dir));
      n2 = norm(diff_ang2(statuses[eid].bcn_dir, statuses[v[j]].bcn_dir));
      if (da >= 0.0) {
        if (n1 > n2) {
          swap_int(&v[i], &v[j]);
        }
      } else {
        if (n1 < n2) {
          swap_int(&v[i], &v[j]);
        }
      }
    }
  }

  for (i = 0; i < nmov; i++) {
    if (id == v[i]) {
      if (da > 0.0) {
        *a = ang[i];
        return;
      } else {
        *a = -ang[i];
        return;
      }
    }
  }
}

void on_ground(struct t_status *s,
               struct beaconMeasure bcn)
{
  static int lid, eid;
  static float ang, orth;

  switch (s->m_fsm) {
    case ONGROUND_ROT :

      /* robot is oriented to the beacon with enough precision */
      if (fabs(bcn.beaconDir) < 4.0) {
        s->lPow = 0.0;
        s->rPow = 0.0;
        s->m_fsm = ONGROUND_GTB;
      } else {
        adjust_angle(s, bcn.beaconDir, 0.0, 0.0, 0.001, 0.0);
      }
      break;

    case ONGROUND_GTB :
      adjust_angle(s, bcn.beaconDir, 0.0, 0.02, 0.001, 0.0);
      if (fabs(bcn.beaconDir) > 90.0 || s->sx > 3.5 || s->cx > 3.5 || s->dx > 3.5 || s->hit) {
        s->m_fsm = ONGROUND_STEADY;
				s->mask |= 0x01 << (id + 16);
				s->mask |= CENTER_FOUND;
				if (get_leader_id(s) == -1) {
          s->mask |= LEADER;
          s->mask |= LEADER_FOUND;
        }
      }
      break;

    case ONGROUND_STEADY :
      lid = get_leader_id(s);
      s->lPow = 0.0;
      s->rPow = 0.0;
      if (lid != id && lid != -1) s->m_fsm = ONGROUND_ORTH;
      break;

    case ONGROUND_ORTH :
      orth = fabs(norm(bcn.beaconDir) - 180.0);
      if (orth < 12.0) {
        s->lPow = 0.0;
        s->rPow = 0.0;
        s->m_fsm = ONGROUND_ORIENTED;
      } else {
        adjust_angle(s, orth, 0.0, 0.0, 0.001, 10.0);
      }
      break;

    case ONGROUND_ORIENTED :
      s->lPow = 0.0;
      s->rPow = 0.0;
      eid = aligned_id(s);
      if (eid != -1) {
        get_final_ang(s, &s->ang_disp);
        float h1, h2, h3;
        h1 = statuses[id].bcn_dir;
        h2 = statuses[eid].bcn_dir + s->ang_disp;
        h3 = diff_ang2(h2, h1);
        if (h3 < 0.0)
          ang = 90.0;
        else
          ang = -90.0;
        s->m_fsm = ONGROUND_ROT2;
      }      
      bcn_dir.d[bcn_dir.i] = bcn.beaconDir;   /* store beacon directions */
      bcn_dir.d1[bcn_dir.i] = get_dir(s);         /* store GPS directions */
      bcn_dir.i++;
      break;

    case ONGROUND_ROT2 :
      orth = fmod(bcn.beaconDir - ang - 360.0, 360.0);
      if (fabs(orth) < 8.0) {
        s->lPow = 0.0;
        s->rPow = 0.0;
        s->m_fsm = ONGROUND_MAKESPACE;
      } else {
        adjust_angle(s, orth, 0.0, 0.0, 0.0005, 0.0);
      }
      break;

    case ONGROUND_MAKESPACE :
      /* should check collisions and the final angular displacement,
			 * but it works fine as it is */
      if (s->cx > 3.5) {
        s->lPow = 0.0;
        s->rPow = 0.0;
      } else
        adjust_angle(s, norm(bcn.beaconDir), norm(ang), 0.01, 0.003, 0.0);
      break;

    case ONGROUND_STOP :
      s->lPow = 0.0;
      s->rPow = 0.0;
      break;
  }
}

void adapt_speed(struct t_status *s)
{
  static int save_sm = -1;

  if (s->t > key_time / 2) sm = 2;
  if (s->t > 3 * key_time / 4) sm = 3;
  if (s->t > key_time) sm = 4;

  if (is_proximity(s, 4.0)) {
    sm = 2;
    save_sm = sm;
  } else {
    if (save_sm != -1) {
      sm = save_sm;
      save_sm = -1;
    }
  }
}

void adapt_speed_old(struct t_status *s)
{
  static int save_sm = -1;

/*
  if (s->t > 1500) sm = 3;
  if (s->t > 2500) sm = 4;
  if (s->t > 4000) sm = 5;
  if (s->t > 6000) sm = 6;
*/
  if (is_proximity(s, 4.0)) {
    sm = 2;
    save_sm = sm;
  } else {
    if (save_sm != -1) {
      sm = save_sm;
      save_sm = -1;
    }
  }
}

/* returns the most precise beacon dir (when dir is steady!!) */
float get_abs_beacon_dir(struct t_status *s)
{
  int i;
  float min1, avg1;
  float min2, avg2;
  float tot;

  if (bcn_dir.i <= 0) {
    tot = norm(bcn.beaconDir + get_dir(s));
    return tot;
  }

  min1 = norm(bcn_dir.d[bcn_dir.i - 1] - 90.0);
  min2 = norm(bcn_dir.d1[bcn_dir.i - 1] - 90.0);
  avg1 = 0.0;
  avg2 = 0.0;
  for (i = 0; i < bcn_dir.i; i++) {
    if (bcn_dir.d[i] > min1) avg1 += bcn_dir.d[i] - 360.0;
    else avg1 += bcn_dir.d[i];
    if (bcn_dir.d1[i] > min2) avg2 += bcn_dir.d1[i] - 360.0;
    else avg2 += bcn_dir.d1[i];
  }  
  avg1 /= bcn_dir.i;  
  avg2 /= bcn_dir.i;  

  tot = fmod(avg1 + avg2 + 720.0, 360.0);

  return tot;
}

int get_closer(struct t_status *s, float *ma)
{
  int i, mid, tid;
  float a1, a2, diff;

  *ma = 720.0;

  tid = get_try_entering_id(s);
  if (tid == -1)
    return -1;

  for (i = 0; i < NROBOT; i++) {
    if ((i != tid) && get_leader_id(s) != i) {
      a1 = statuses[tid].bcn_dir;
      a2 = statuses[i].bcn_dir;
      diff = diff_ang(a1, a2);
      if (diff < *ma) {
        *ma = diff;
        mid = i;
      }
    }
  }
  return mid;
}

int better_to_finish(struct t_status *s)
{
  int dt, n, gain, gainable, over;

  dt = final_time - key_time;
  over = s->t - key_time;
  n = how_many_onground();
  gain = n * (int)(100.0 * (float)(dt - over) / (float)(dt));
  gainable = (NROBOT - n) * (int)(100.0 * (float)(dt - over) / (float)(dt));

  if (over < 0)
    return 0;

  if (gain > gainable) return 1;

  if (s->t > key_time + (dt / 2)) {
    return 1;
  }
  return 0;
}

/* manage the FSM */
void fsm(struct t_status *s, int beaconToFollow)
{
  static float d;
  static int save_fsm;
  static int tmp, tmp1;
  static int timer1 = -1, timer2 = -1;

  /* access to values from Sensors - Only ReadSensors() gets new values */
  if (IsObstacleReady(LEFT))
    s->sx = GetObstacleSensor(LEFT);
  if (IsObstacleReady(RIGHT))
    s->dx = GetObstacleSensor(RIGHT);

  s->bcn_rdy = IsBeaconReady(beaconToFollow);
  if (s->bcn_rdy) {
    bcn =  GetBeaconSensor(beaconToFollow);
    s->bcn_dir = bcn.beaconDir;
    s->bcn_vis = bcn.beaconVisible;
  } else {
    s->bcn_rdy = 0;
  }

  if (s->bcn_rdy && bcn.beaconVisible) { 
	  beacon_found(s, bcn.beaconDir);
  }

  if (IsBumperReady()) s->hit = GetBumperSensor();
  if (IsObstacleReady(CENTER)) s->cx = GetObstacleSensor(CENTER);

  d = dist(s->goalx, s->goaly, s->x, s->y);

  add_dist_measure(d);

  tmp1 = get_leader_id(s);
  if (tmp1 == -1) {
    tmp = check_someone_finding_center(s);
    if (tmp && s->fsm != WAIT) {
      save_fsm = s->fsm;
      s->fsm = WAIT;
      timer2 = s->t;
    }
  }

  switch (s->fsm) {
    case QUEST :
      if (check_try_enter(s)) {
        s->fsm = TRY_ENTER;
        break;
      }
      if (s->gnd) {
        s->m_fsm = ONGROUND_ROT;
        s->fsm = ON_GROUND;
        break;
      }
      if (s->bcn_rdy && bcn.beaconVisible) { 
        new_motion2goal(s, DIR_FOUND);
        break;
      }
      quest(s, d, bcn);
      break;

    case DIR_FOUND :
		if (check_try_enter(s)) {
        s->fsm = TRY_ENTER;
        break;
      }
      if (s->gnd) {
        s->m_fsm = ONGROUND_ROT;
        s->fsm = ON_GROUND;
        break;
      }
      dir_found(s, d, bcn);
      break;

    case GROUND_FOUND :
      if (check_try_enter(s)) {
        s->fsm = TRY_ENTER;
        break;
      }
      if (s->gnd) {
        s->m_fsm = ONGROUND_ROT;
        s->fsm = ON_GROUND;
        break;
      }
      ground_found(s, d, bcn);
      break;

    case ON_GROUND :
      on_ground(s, bcn);
      break;

    case WAIT :
      tmp1 = get_leader_id(s);
      if (tmp1 != -1) {
        s->fsm = save_fsm;        
        break;
      }
      /* this is a quick fix for some robots that hang in this state
			 * when a node is being elected as leader... to be fixed in
			 * some other way... */
      if (s->t - timer2 > 100) {
        s->fsm = save_fsm;
        break;
      }
      s->lPow = 0.0;
      s->rPow = 0.0;
      break;

    case TRY_ENTER : /* rotate towards the beacon */
      if (timer1 < 0) {
        timer1 = s->t;
      }
      if (s->gnd) {
        s->fsm = ON_GROUND;
        break;
      }
      if (fabs(bcn.beaconDir) < 8.0) {
        if (s->sx > 3.5 || s->cx > 3.5 || s->dx > 3.5 || s->hit) {
          s->lPow = -0.01;
          s->rPow = -0.01;
        } else {
          s->lPow = 0.01;
          s->rPow = 0.01;
        }
        s->fsm = TRY_ENTER;
        bcn_dir.d[bcn_dir.i] = bcn.beaconDir;   /* store beacon directions */
        bcn_dir.d1[bcn_dir.i] = get_dir(s);         /* store GPS directions */
        bcn_dir.i++;

        /* gives time to filter the angle; larger delay means more
				 * latency but more filtering (of course!) */
        if (s->t - timer1 > 50) {
          s->fsm = TRY_ENTER_ALIGNED;
        }
      } else {
        /* turn to the beacon's direction */
        adjust_angle(s, bcn.beaconDir, 0.0, 0.0, 0.001, 0.0);
      }
      break;

    case TRY_ENTER_ALIGNED :
      if (s->gnd) {
        s->lPow = 0.01;
        s->rPow = 0.01;
        s->fsm = ON_GROUND;
        s->m_fsm = ONGROUND_STOP;
      }

      if (fabs(bcn.beaconDir) < 8.0) {
        if (s->sx > 5.5 || s->cx > 5.5 || s->dx > 5.5 || s->hit) {
          s->lPow = -0.01;
          s->rPow = -0.01;
        } else {
          s->lPow = 0.01;
          s->rPow = 0.01;
        }
        s->fsm = TRY_ENTER_ALIGNED;
        bcn_dir.d[bcn_dir.i] = bcn.beaconDir;   /* store beacon directions */
        bcn_dir.d1[bcn_dir.i] = get_dir(s);         /* store GPS directions */
        bcn_dir.i++;
      } else {
        /* turn to the beacon's direction */
        adjust_angle(s, bcn.beaconDir, 0.0, 0.0, 0.001, 0.0);
      }
      break;
  }

  adapt_speed(s);

  if (better_to_finish(s) && (s->fsm == ON_GROUND)) {
    fprintf(outf, "robot %d: is better to finish!\n", id);
    s->fsm = CAN_FINISH;
  }

  if (how_many_onground() == NROBOT) {
    fprintf(outf, "robot %d: done!\n", id);
    s->fsm = CAN_FINISH;
  }
}

void set_msg(struct t_status *s)
{
  float dir, xgnd, ygnd;

  if ((s->fsm == ON_GROUND &&
      (s->m_fsm == ONGROUND_ORTH ||
       s->m_fsm == ONGROUND_ORIENTED)) || 
       s->fsm == TRY_ENTER_ALIGNED)
    dir = get_abs_beacon_dir(s);
  else
    dir = bcn.beaconDir + get_dir(s);

  if (s->mask & LEADER) {
    get_pos_center(50, &xgnd, &ygnd);
  } else {
    xgnd = s->xgnd;
    ygnd = s->ygnd;
  }

  sprintf(s->msg,
          "%d %x %2d %2d %4.1f %4.1f %4.1f %4.1f %5.1f %4.1f %4.1f %4.1f %d",
           id, s->mask, s->fsm, s->m_fsm, 
           s->x, s->y, s->xc, s->yc, s->dir2bcn,
           xgnd, ygnd, dir, s->gnd);
}

/*
 * The prio level is used to keep track of the accuracy of the
 * ground co-ordinates. Higher prio means that the x/ygnd values
 * have been set starting from higher precision information.
 */
void manage_comm(struct t_status *s)
{
  int i;
  int mid;
  char *retmsg;
  char prio = 0;

  set_msg(s);
  Say(s->msg);

  for (i = 0; i < NROBOT; i++) {
    if ((i != id) && NewMessageFrom(i)) {
      retmsg = GetMessageFrom(i);
      sscanf(retmsg, "%d %x %d %d %f %f %f %f %f %f %f %f %d",
             &mid,
             &statuses[i].mask,
             &statuses[i].fsm,
             &statuses[i].m_fsm,
             &statuses[i].x,
             &statuses[i].y,
             &statuses[i].xc, &statuses[i].yc, &statuses[i].dir2bcn, 
             &statuses[i].xgnd, &statuses[i].ygnd,
             &statuses[i].bcn_dir,
             (int *)&statuses[i].gnd);
      s->mask |= statuses[i].mask;
      if (statuses[i].mask & BEACON) {
        if (s->record_new_beacon) {
        s->mask |= statuses[i].mask;
				s->mask |= 0x01 << id;
        s->xc = statuses[i].xc;
        s->yc = statuses[i].yc;
        s->dir2bcn = statuses[i].dir2bcn;
        prio = 1;
        }
      }
      if (statuses[i].mask & TARGET) {
        s->mask |= statuses[i].mask;
				s->mask |= 0x01 << (id + 8);
        if (prio < 2) {
          s->xgnd = statuses[i].xgnd;
          s->ygnd = statuses[i].ygnd;
          prio = 2;
        }
      }
      if (statuses[i].mask & CENTER_FOUND) {
        s->mask |= statuses[i].mask;
				s->mask |= 0x01 << (id + 16);
        if (prio < 3) {
          s->xgnd = statuses[i].xgnd;
          s->ygnd = statuses[i].ygnd;
          prio = 3;
        }
      }
      if (statuses[i].mask & LEADER) {
        s->mask &= ~LEADER;   /* zeros the leader flag */
        if (prio < 4) {
          s->xgnd = statuses[i].xgnd;
          s->ygnd = statuses[i].ygnd;
          prio = 4;
        }
      }
    }
  }
  s->bcn_dir = get_abs_beacon_dir(s);
  statuses[id] = *s;
}

void init_status(struct t_status *s)
{

  dest_id = init_dest_id[id];
  new_motion2goal(s, QUEST);   /* init s->fsm; s->m_fsm; */
  s->msg = malloc(1000);
  s->msg[0] = '\0';
  s->record_new_beacon = 1;
  s->mask = 0;
  s->idir = 0;
  s->full = 0;
  s->gnd = 0;
}

void main_loop() {
  int state = STOP, stoppedState = RUN;
  int beaconToFollow = 0;
  int gndrdy, gnd;
  struct t_status status;
  unsigned int t;

  state = STOP;

  init_status(&status);

  while (1) {

    /* read next values from sensors */
    ReadSensors();

    /* simulator has received Finish() or robot removed */
    if (GetFinished()) {
      fprintf(outf, "exiting %s\n", name);
      exit(0);
    }

    /* restart */
    if (state == STOP && GetStartButton())
      state = stoppedState;

    if (state != STOP && GetStopButton()) {
      stoppedState = state;
      state = STOP;   /* interrupt */
    }
  
    if (status.fsm == CAN_FINISH) {
      Finish();
      fprintf(outf, "%s on target %d at %d\n", name, beaconToFollow, GetTime());
    }

    if (state == RUN) { 
      gndrdy = IsGroundReady();
      if (gndrdy) {
        gnd = (GetGroundSensor() == beaconToFollow);
        if (gnd) {         /* on target */
          status.mask |= TARGET;
	  			status.mask |= 0x01 << (id + 8);
          status.xgnd = status.x;
          status.ygnd = status.y;
        }
        status.gnd = gnd;
      }
	  }

    get_gps(&status);
    add_pos_measure(status.x, status.y);

    if (state == RUN) {
      fsm(&status, beaconToFollow);
      if (!isnan(status.lPow) && !isnan(status.rPow))
        DriveMotors(status.lPow, status.rPow);
      manage_comm(&status);
    }

	  /* request sensors for next cycle */
    t = GetTime();
	  if ((t % 2) == 0) {
      RequestObstacleSensor(CENTER);

      if ((t % 8) == 0 || beaconToFollow == GetNumberOfBeacons()) {
        RequestGroundSensor();
      } else {
        RequestBeaconSensor(beaconToFollow);
      }
    } else {
      RequestSensors(2, "IRSensor1", "IRSensor2");
    }
  }
}

int main(int argc, char *argv[])
{
  char host[100] = "localhost";

  printf(" t-bots\n Copyright (C) 2008 University of Pavia\n" );

  /* processing arguments */
  /* every option has a value, thus argc must be 1, 3, 5, ... */
  while (argc > 2) {
    if (strcmp(argv[1], "-host") == 0) {
      strncpy(host, argv[2], 99);
      host[99] = '\0';
    } else if (strcmp(argv[1], "-robname") == 0) {
      strncpy(name, argv[2], 19);
      name[19] = '\0';
    } else if (strcmp(argv[1], "-pos") == 0) {
      if (sscanf(argv[2], "%d", &id) != 1)
        argc = 0;   /* error message will be printed */
    } else {
      break;   /* while */
    }
    argc -= 2;
    argv += 2;
  }

  if (argc != 1) {
    fprintf(stderr, "bad number of parameters\n"
                    "SYNOPSIS: mainRob [-host hostname] [-robname robotname] [-pos posnumber]\n");
    return 1;
  }
  
  /* Connect Robot to simulator */
  if (InitRobot(name, id, host) == -1) {
    printf("%s Failed to connect\n", name); 
    exit(1);
  }

  id--;

  printf("%s Connected\n", name);
  final_time = GetFinalTime();
  key_time = GetKeyTime();

  pos_hist.i = 0;
  bcn_dir.i = 0;

  sprintf(fname, "bot%d.log", id);
  outf = stdout;
//  outf = fopen(fname, "w");

  main_loop();

  fclose(outf);

  return 1;
}
