#ifndef __tbots_h__
#define __tbots_h__

#include "RobSock.h"

/* used for debugging: enable print out from given robot(s) */
#define PRINT_BOT (id == 0)
//#define PRINT_BOT (1)

#define TMAX  (20000)

#define DIST_INFTY   (10000.0)

/* simulation status related to simulator's button inputs  */
#define RUN (1)
#define STOP (2)
#define FINISHED (4)

#define NROBOT (5)

#define NDIRS  (1)

/* size of distance history (larger values gives more filtering) */
#define DIST_HIST_MAX (30)

/* motion status for QUEST, DIR_FOUND, GROUND_FOUND */
#define MOTION_TO_GOAL     (0)
#define BOUNDARY_FOLLOWING (1)
#define ON_GOAL            (2)

/* beacon's search */
#define QUEST        (0)

/* _DIR_FOUNDs define the statuses during the "go-to-beacon" phase */
#define DIR_FOUND    (1)

/* robot received the information about the exact location of the
 * target and goes there
 */
#define GROUND_FOUND (2)

/* the robot is on the target, but doesn't finish until it
 * communicate its location to all other robots */
#define ON_GROUND    (3)

/* rotating towards the beacon */
#define ONGROUND_ROT    (10)
/* going towards the beacon */
#define ONGROUND_GTB    (11)
/* center of the  */
#define ONGROUND_STEADY (12)
#define ONGROUND_ORTH   (13)
#define ONGROUND_ORIENTED   (14)
#define ONGROUND_ROT2   (15)
#define ONGROUND_MAKESPACE   (16)
#define ONGROUND_STOP   (17)

#define WAIT (20)

#define TRY_ENTER  (50)
#define TRY_ENTER_ALIGNED  (51)

/* the robot is on the target, and received confirmation that all
* friends received its positions: now can finish! */
#define CAN_FINISH   (-1)

/* found the beacon */
#define BEACON         (0x01000000)
/* over the target area */
#define TARGET         (0x02000000)
#define CENTER_FOUND   (0x04000000)
#define LEADER_FOUND   (0x08000000)
#define LEADER         (0x10000000)

/*
 * We use a regression of previous distances to determine if we are
 * decreasing the distance to the goal since, due to GPS accuracy
 * errors, we can not assess it by only comparing distances between
 * two consecutive time instants (distance may erroneously increase
 * due to GPS accuracy errors).
 */

struct t_dist_hist {
  float dist[DIST_HIST_MAX];
  int i;
  int full;
};

struct t_pos_hist {
  float x[TMAX];
  float y[TMAX];
  int i;
};

struct t_bcn_dir_hist {
  float d[TMAX];   /* dir to beacon */
  float d1[TMAX];  /* gps dir */
  int i;
};

/*
 * Statuses contain into MSB the code of operation/event
 * LSB contain the agreement bits, one for each robot.
 * Messages are in the form: status params ... status params ...
 */
struct t_status {
  char record_new_beacon;
  int mask;
  int fsm;   /* contains the current FSM's state */
  int m_fsm;   /* contains the current mobility state */
  float xc, yc;   /* location where the beacon was received */
  float dir2bcn;   /* heading to beacon when in (xc, yc) */
  float xgnd, ygnd;   /* target co-ordinates, on ground */
  float sx, cx, dx;   /* proximity sensor's values */
  bool hit;   /* bumper sensor value */
  float lPow, rPow;   /* motor control power */
  float goalx, goaly;   /* current destination */
  bool gnd;   /* on ground flag */

  float ang_disp;   /* final angular displacement to make space */
  /* errors when putting ang_disp (float with THIS NAME) below here */

  /* GPS data */
  bool gps_rdy;
  float x;
  float y;
  float dir[NDIRS];
  char full;   /* the dir sample is fullfilled */
  int idir;    /* index of the next sample to insert */
  int t;

  /* beacon data */
  bool bcn_rdy;   /* beacon ready */
  float bcn_dir;
  bool bcn_vis;
  char *msg;
};

/* 
 * contains info useful to the boundary following procedure;
 * contains the starting co-ordinates and time at which the boundary
 * following procedure had started.
 */
struct t_bf_info {
  float x;  
  float y;
  long t;
};

struct t_speed {
  float max;
  int inf_loop_timer;
  int pos_recurr_timer;
};

void new_motion2goal(struct t_status *s, int next);
int is_proximity(struct t_status *s, float dth);
int how_many_onground();
int get_leader_id(struct t_status *s);
int get_try_entering_id(struct t_status *s);
int get_closer(struct t_status *s, float *ma);
float get_abs_beacon_dir(struct t_status *s);

#endif
